/*
 * Copyright (c) 2010, Stephen Denne
 */
package nz.co.datacute.pgquilt;

import java.net.URL;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import nz.co.datacute.treemap.Node;
import nz.co.datacute.treemap.NodePositioner;
import nz.co.datacute.treemap.NodeRenderer;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Rect;
import android.graphics.drawable.BitmapDrawable;
import android.os.AsyncTask;
import android.preference.PreferenceManager;
import android.util.AttributeSet;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnKeyListener;
import android.widget.TextView;

public class TreeMapView extends View implements OnKeyListener {

	private boolean newState;
	private Node root;
	private boolean needLayout = true;
	private boolean keyPressHandled = false;
	private Node nodeWithFocus = null;
	private Node prevNodeWithFocus = null;
	// Cache of backgrounds, for various weighting and category selections, per orientation
	private Map<String,Bitmap> backgroundCache;
	private ProgressDialog progressDialog;
	
	private static final Paint focusPaint = new Paint();
	static {
		focusPaint.setStyle(Paint.Style.STROKE);
		focusPaint.setColor(Color.WHITE);
		focusPaint.setStrokeWidth(0.0f);
	}

	public TreeMapView(Context context) {
		this(context, null,0);
	}

	public TreeMapView(Context context, AttributeSet attrs) {
		this(context, attrs,0);
	}

	public TreeMapView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		initLayout();
	}

	private void initLayout() {
		setOnKeyListener(this);
		setBackgroundDrawable(null);
		setWillNotDraw(false);
	}

	@Override
	protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        final int widthMode = MeasureSpec.getMode(widthMeasureSpec);
        final int widthSize = MeasureSpec.getSize(widthMeasureSpec);

        final int heightMode = MeasureSpec.getMode(heightMeasureSpec);
        final int heightSize = MeasureSpec.getSize(heightMeasureSpec);

		if (widthMode != MeasureSpec.EXACTLY || heightMode != MeasureSpec.EXACTLY) {
            throw new IllegalStateException("NodeView can only be used with MeasureSpec mode=EXACTLY");
        }

        setMeasuredDimension(widthSize, heightSize);
	}
	
	@Override
	protected void onLayout(boolean changed, int l, int t, int r, int b) {
		if (needLayout && (root != null)) {
			NodePositioner.layout(root, r-l-1, b-t-1);
			needLayout = false;
			invalidate();
		}
	}
	
	@Override
	protected void onDraw(Canvas canvas) {
		if (root == null) {
			String refreshing = getContext().getString(R.string.dialog_message_refreshing);
			if ((progressDialog != null) && (progressDialog.isShowing())) {
				progressDialog.setMessage(refreshing);
				progressDialog.show();
			} else {
				progressDialog = ProgressDialog.show(getContext(), null, refreshing, true);
			}
		} else {
			if (newState) {
				newState = false;
				updateBackground();
			}
			if (getBackground() == null) {
				new BackgroundRenderer().execute(root);
				String rendering = getContext().getString(R.string.dialog_message_rendering);
				if ((progressDialog != null) && (progressDialog.isShowing())) {
					progressDialog.setMessage(rendering);
					progressDialog.show();
				} else {
					progressDialog = ProgressDialog.show(getContext(), null, rendering, true);
				}
			} else {
				if ((progressDialog != null) && (progressDialog.isShowing())) {
					progressDialog.dismiss();
				}
				if (nodeWithFocus == null) {
					setFocusToFirstChild();
				}
				canvas.drawRect(nodeWithFocus.x, nodeWithFocus.y, nodeWithFocus.x + nodeWithFocus.width, nodeWithFocus.y + nodeWithFocus.height, focusPaint);
			}
		}
	}

	@Override
	public boolean onTouchEvent(MotionEvent event) {
        int x = (int) event.getX();
        int y = (int) event.getY();
        Rect r = new Rect();
        // Check current focus first
		if (nodeWithFocus != null) {
    		r.set(nodeWithFocus.x, nodeWithFocus.y, nodeWithFocus.x + nodeWithFocus.width, nodeWithFocus.y + nodeWithFocus.height);
			if (r.contains(x, y)) {
				return true;
			}
		}
//		nodeWithFocus = null;
		if (root != null) {
			focusAt(x, y, r);
			invalidate();
		}
		return true;
	}

	private void focusAt(int x, int y, Rect r) {
        Node parent = root;
        boolean checkChildren = true;
        while (checkChildren) {
        	checkChildren = false;
        	for (Node c : parent.getChildren()) {
        		r.set(c.x, c.y, c.x + c.width, c.y + c.height);
    			if (r.contains(x, y)) {
    				if (c.getChildren().isEmpty()) {
    					if (!c.equals(nodeWithFocus)) {
    						setFocusNode(c);
    					}
    				} else {
    					parent = c;
    					checkChildren = true;
    				}
    				break;
    			}
        	}
        }
	}

	private void setFocusNode(Node c) {
		nodeWithFocus = c;
		updateFocusInfo(c.fullLabel, root.categoryLabels[root.getChosenCategory()] + ": "+ c.getCategory().label, root.weightLabels[root.getChosenWeight()] + ": " + c.getWeightDescription());
	}

	private void updateFocusInfo(String label, String cateory, String weight) {
		View activityView = getRootView();
		((TextView) activityView.findViewById(R.id.selected_item_label)).setText(label);
		((TextView) activityView.findViewById(R.id.selected_item_category)).setText(cateory);
		((TextView) activityView.findViewById(R.id.selected_item_weight)).setText(weight);
	}
	
	private class RootObtainer extends AsyncTask<URL,Void,Node> {
		
		Exception storedException;

		@Override
		protected Node doInBackground(URL... urls) {
			try {
				SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(getContext());
				if (prefs.getBoolean("demo", false)) {
					return DemoCreator.getRoot(getContext());
				} else {
					return SQLCreator.getRoot(getContext());
				}
			} catch (Exception e) {
				e.printStackTrace();
				storedException = e;
			}
			return null;
		}

		@Override
		protected void onPostExecute(Node result) {
			if (result != null) {
				root = result;
				needLayout = true;
				requestLayout();
			} else {
				if ((progressDialog != null) && (progressDialog.isShowing())) {
					progressDialog.dismiss();
				}
				new AlertDialog.Builder(getContext())
					.setMessage(storedException.toString())
					.setCancelable(false)
					.setPositiveButton(R.string.menu_preferences, new DialogInterface.OnClickListener() {
						@Override
						public void onClick(DialogInterface dialog, int which) {
					        View refreshButton = getRootView().findViewById(R.id.refresh);
							refreshButton.setEnabled(true);
							getContext().startActivity(new Intent(getContext(), ConnectionPreferences.class));
						}
					})
					.setNeutralButton(R.string.menu_try_demo, new DialogInterface.OnClickListener() {
						@Override
						public void onClick(DialogInterface dialog, int which) {
							SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(getContext());
							Editor editor = prefs.edit();
							editor.putBoolean("demo", true);
							editor.commit();
					        View refreshButton = getRootView().findViewById(R.id.refresh);
							refreshButton.setEnabled(true);
							refreshButton.performClick();
						}
					})
					.setNegativeButton(R.string.menu_retry, new DialogInterface.OnClickListener() {
						@Override
						public void onClick(DialogInterface dialog, int which) {
					        View refreshButton = getRootView().findViewById(R.id.refresh);
							refreshButton.performClick();
						}
					})
					.show();
				storedException = null;
			}
		}
	}

	private boolean replaceFocus(Node parent) {
		if (prevNodeWithFocus == null) {
			setFocusToFirstChild();
			return true;
		}
		List<Node> children = parent.getChildren();
		if (children.isEmpty()) {
			if (parent.fullLabel.equals(prevNodeWithFocus.fullLabel) && parent.getCategory().equals(prevNodeWithFocus.getCategory())) {
				setFocusNode(parent);
				return true;
			}
		} else {
			for (Node child : children) {
    			if (replaceFocus(child)) {
    				return true;
    			}
			}
		}
        return false;
	}
	
	private class BackgroundRenderer extends AsyncTask<Node,Void,Bitmap> {

		@Override
		protected Bitmap doInBackground(Node... nodes) {
			Bitmap bitmap = Bitmap.createBitmap(nodes[0].width + 1, nodes[0].height + 1, Bitmap.Config.ARGB_8888);
			NodeRenderer.render(nodes[0], new Canvas(bitmap));
			return bitmap;
		}

		@Override
		protected void onPostExecute(Bitmap result) {
			if (nodeWithFocus == null) {
				replaceFocus(root);
			}
			if (backgroundCache == null) {
				backgroundCache = new HashMap<String, Bitmap>();//new Bitmap[root.getAvailableWeightings()][root.getAvailableCategories()];
			}
			backgroundCache.put(root.cacheKey(),result);//[root.getChosenWeight()][root.getChosenCategory()] = result;
			setBackgroundDrawable(new BitmapDrawable(result));
	        View refreshButton = getRootView().findViewById(R.id.refresh);
			refreshButton.setEnabled(true);
		}
	}

	public void refresh() {
		root = null;
		needLayout = true;
		prevNodeWithFocus = nodeWithFocus;
		nodeWithFocus = null;
		backgroundCache = null;
		setBackgroundDrawable(null);
		updateFocusInfo("","","");
		new RootObtainer().execute((URL[])null);
	}

	public void changeWeighting(int chosenWeighting) {
		root.setChosenWeight(chosenWeighting);
		needLayout = true;
		updateBackground();
		requestLayout();
	}

	public void changeCategory(int chosenCategory) {
		root.setChosenCategory(chosenCategory);
		updateBackground();
		invalidate();
	}

	private void updateBackground() {
		setFocusNode(nodeWithFocus);
		if (backgroundCache == null) {
			backgroundCache = new HashMap<String, Bitmap>();//new Bitmap[root.getAvailableWeightings()][root.getAvailableCategories()];
		}
		Bitmap bitmap = backgroundCache.get(root.cacheKey());//[root.getChosenWeight()][root.getChosenCategory()];
		if (bitmap != null) {
			setBackgroundDrawable(new BitmapDrawable(bitmap));
		} else {
			setBackgroundDrawable(null);
		}
	}

	@Override
	public boolean onKey(View v, int keyCode, KeyEvent event) {
		if (keyPressHandled && (event.getAction() == KeyEvent.ACTION_UP)) {
			keyPressHandled = false;
			return true;
		}
		switch (keyCode) {
		case KeyEvent.KEYCODE_DPAD_UP:
			if (nodeWithFocus == null) {
				setFocusToFirstChild();
			} else {
				focusAt(nodeWithFocus.x, nodeWithFocus.y - 1, new Rect());
			}
			invalidate();
			keyPressHandled = true;
			return true;
		case KeyEvent.KEYCODE_DPAD_DOWN:
			if (nodeWithFocus == null) {
				setFocusToFirstChild();
			} else {
				focusAt(nodeWithFocus.x, nodeWithFocus.y + nodeWithFocus.height, new Rect());
			}
			invalidate();
			keyPressHandled = true;
			return true;
		case KeyEvent.KEYCODE_DPAD_LEFT:
			if (nodeWithFocus == null) {
				setFocusToFirstChild();
			} else {
				focusAt(nodeWithFocus.x - 1, nodeWithFocus.y, new Rect());
			}
			invalidate();
			keyPressHandled = true;
			return true;
		case KeyEvent.KEYCODE_DPAD_RIGHT:
			if (nodeWithFocus == null) {
				setFocusToFirstChild();
			} else {
				focusAt(nodeWithFocus.x + nodeWithFocus.width, nodeWithFocus.y, new Rect());
			}
			invalidate();
			keyPressHandled = true;
			return true;
		}
		return false;
	}

	private void setFocusToFirstChild() {
		Node child = root;
		while (!child.getChildren().isEmpty()) {
			Node[] children = child.getChildren().toArray(new Node[0]);
			Arrays.sort(children);
			child = children[0];
		}
		setFocusNode(child);
	}

	public Node getRoot() {
		return root;
	}

	public void setRoot(Node root) {
		this.root= root;
		needLayout = true;
		requestLayout();
	}

	public void onLowMemory() {
		backgroundCache = null;
	}
	
	public Object getState() {
		State state = new State();
		state.root = root;
		state.nodeWithFocus = nodeWithFocus;
		state.bitmapCache = backgroundCache;
		return state;
	}

	public void setState(Object lastNonConfigurationInstance) {
		if (lastNonConfigurationInstance instanceof State) {
			State state = (State) lastNonConfigurationInstance;
			root = state.root;
			backgroundCache = state.bitmapCache;
			nodeWithFocus = state.nodeWithFocus;
			newState = true;
			needLayout = true;
		}
		if (root == null) {
			new RootObtainer().execute((URL[])null);
		}
	}
	
	private static class State {
		Node root;
		Node nodeWithFocus = null;
		private Map<String,Bitmap> bitmapCache;
	}
}
