/*
 * Copyright (c) 2010, Stephen Denne
 */
package nz.co.datacute.treemap;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

public class Node implements Serializable, Comparable<Node> {
	
	private static final long serialVersionUID = 3120557677073061087L;

	public final String parent_oid;
	public final String oid;
	public final String fullLabel;
	public final String shortLabel;

	public Node parent = null;
	final List<Node> children;

	private int chosenWeight = 0;
	private double[] weights;
	public String[] weightDescriptions;
	public String[] weightLabels;
	private int chosenCategory = 0;
	private NodeCategory[] categories;
	public String[] categoryLabels;

	// Laid out rectangle location and size
	public int x;
	public int y;
	public int width;
	public int height;
	
	public Node(String fullLabel, String shortLabel) {
		this.parent_oid = null;
		this.oid = null;
		this.fullLabel = fullLabel;
		this.shortLabel = shortLabel;
		this.children = new LinkedList<Node>();
	}
	
	public Node(String fullLabel, String shortLabel, String[] weightLabels, String[] categoryLabels) {
		this.parent_oid = null;
		this.oid = null;
		this.fullLabel = fullLabel;
		this.shortLabel = shortLabel;
		this.children = new LinkedList<Node>();
		this.weights = new double[weightLabels.length];
		this.weightLabels = weightLabels;
		this.categoryLabels = categoryLabels;
	}
	
	public Node(String fullLabel, String shortLabel, double[] weights, NodeCategory[] categories) {
		this.parent_oid = null;
		this.oid = null;
		this.fullLabel = fullLabel;
		this.shortLabel = shortLabel;
		this.weights = weights;
		this.categories = categories;
		this.children = Collections.emptyList();
	}

	public Node(String fullLabel, String shortLabel, double[] weights, String[] weightDescriptions, NodeCategory[] categories) {
		this(null, fullLabel, fullLabel, shortLabel, weights, weightDescriptions, categories);
	}

	public Node(String parent_oid, String oid, String fullLabel, String shortLabel, double[] weights, String[] weightDescriptions, NodeCategory[] categories) {
		this.parent_oid = parent_oid;
		this.oid = oid;
		this.fullLabel = fullLabel;
		this.shortLabel = shortLabel;
		this.weights = weights;
		this.categories = categories;
		this.children = new ArrayList<Node>();
		this.weightDescriptions = weightDescriptions;
	}

	public Node add(Node node) {
		children.add(node);
		node.parent = this;
		addWeights(node);
		return this;
	}
	
	public void addWeights(Node node) {
		if ((weights == null) && (node.weights != null)) {
			weights = new double[node.weights.length];
//			weightDescriptions = new String[node.weights.length];
		}
		for (int i = 0; i < weights.length; i++) {
			weights[i] += node.weights[i];
		}
		if (parent != null) {
			parent.addWeights(this);
		}
	}
	
	public String describeLayout() {
		StringBuilder sb = new StringBuilder();
		describeLayout(sb);
		return sb.toString();
	}

	public void describeLayout(StringBuilder sb) {
		sb.append(fullLabel);
		sb.append(" = ");
		sb.append(getWeight());
		sb.append(" [");
		sb.append(width);
		sb.append(" x ");
		sb.append(height);
		sb.append("] (");
		sb.append(width * height);
		sb.append(") @ ");
		sb.append(x);
		sb.append(",");
		sb.append(y);
		if (!children.isEmpty()) {
			sb.append(" {\n");
			Node[] childArray = children.toArray(new Node[0]);
			Arrays.sort(childArray);
			for (Node child : childArray) {
				child.describeLayout(sb);
			}
			sb.append("}");
		}
		sb.append("\n");
	}

	@Override
	public int compareTo(Node o) {
		if (getWeight() < o.getWeight()) return 1;
		if (getWeight() > o.getWeight()) return -1;
		return shortLabel.compareTo(o.shortLabel);
	}

	public List<Node> getChildren() {
		return children;
	}

	public double getWeight() {
		return weights[chosenWeight];
	}

	public String getWeightDescription() {
		String desc = weightDescriptions[chosenWeight];
		return desc == null ? String.format("%-1.0f",weights[chosenWeight]) : desc;
	}
	
	public int getChosenWeight() {
		return chosenWeight;
	}
	
	public void setChosenWeight(int chosenWeight) {
		this.chosenWeight = chosenWeight;
		for (Node child : children) {
			child.setChosenWeight(chosenWeight);
		}
	}
	
	public int getAvailableWeightings() {
		return weightLabels.length;
	}

	public NodeCategory getCategory() {
		if (categories == null) return parent.getCategory();
		if (categories[chosenCategory] == null) return parent.getCategory();
		return categories[chosenCategory];
	}

	public int getChosenCategory() {
		return chosenCategory;
	}
	
	public void setChosenCategory(int chosenCategory) {
		this.chosenCategory = chosenCategory;
		for (Node child : children) {
			child.setChosenCategory(chosenCategory);
		}
	}
	
	public int getAvailableCategories() {
		return categoryLabels.length;
	}

	public String cacheKey() {
		return String.valueOf(width) + "x" + String.valueOf(height) + "," + String.valueOf(chosenWeight) + "," + String.valueOf(chosenCategory);
	}
}
