/*
 * Copyright (c) 2010, Stephen Denne
 */
package nz.co.datacute.treemap;

import java.util.List;

import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.RadialGradient;
import android.graphics.Rect;
import android.graphics.Shader;
import android.graphics.Paint.FontMetrics;

public class NodeRenderer {

	private static final Paint fillPaint = new Paint();
	private static final Paint textPaint = new Paint(Paint.ANTI_ALIAS_FLAG | Paint.SUBPIXEL_TEXT_FLAG | Paint.DEV_KERN_TEXT_FLAG);
	private static final Rect stringBounds = new Rect();
	private static final FontMetrics fontMetrics = new FontMetrics();
	static {
		fillPaint.setStyle(Paint.Style.FILL);
		fillPaint.setStrokeWidth(0.0f);
		textPaint.setStyle(Paint.Style.STROKE);
		textPaint.setColor(Color.WHITE);
		textPaint.setTextSize(12.0f);
		textPaint.setTextAlign(Paint.Align.LEFT);
	}

	
	public static Canvas render(Node node, Canvas canvas) {
		int x = node.x;
		int y = node.y;
		int width = node.width;
		int height = node.height;
		
		if ((width == 0) || (height == 0)) return canvas;
		
		List<Node> children = node.getChildren();

		if (!children.isEmpty()) {
			for (Node child : children) {
				render(child,canvas);
			}
		} else {
			float cx = x + width / 2.0f;
			float cy = y + height / 2.0f;
			float radius = (width + height)/2.0f;
			Node parent = node.parent;
			float weighting = (float)node.getWeight();
			float decay = 0.5f;
			while (parent != null) {
				// adjust toward parents, depending on how big a piece of the parent this is.
				// weighting half of parent's weighting means this is half of parent.
				//    move centre slightly
				// weighting 1/10 of parent's weighting means this is 1/10 of parent.
				//    move centre a lot
				// decay the affect the more generations traversed
				float pcx = parent.x + parent.width / 2.0f;
				float pcy = parent.y + parent.height / 2.0f;
				float pr = (parent.width + parent.height)/2.0f;
				float pw = (float)parent.getWeight();
				float ratio = ((pw - weighting) / pw) * decay;
				cx = cx * (1.0f - ratio) + pcx * ratio;
				cy = cy * (1.0f - ratio) + pcy * ratio;
				radius = radius * (1.0f - ratio) + pr * ratio;
				weighting = pw;
				parent = parent.parent;
				decay *= decay;
			}

			fillPaint.setShader(new RadialGradient(cx, cy, radius * 0.8f, node.getCategory().color, Color.BLACK, Shader.TileMode.CLAMP));
			int paintedX = x;
			if (width>1) {
				paintedX++;
				width--;
			}
			int paintedY = y;
			if (height>1) {
				paintedY++;
				height--;
			}
			canvas.drawRect(paintedX, paintedY, paintedX + width, paintedY + height, fillPaint);
			
			textPaint.setTextSize(12.0f);
			int line = 0;
			if (showLine(node, canvas, textPaint, line, node.shortLabel) == 1) {
				line++;
				line += showLine(node, canvas, textPaint, line, node.getWeightDescription());
			}
		}
		return canvas;
	}

	private static int showLine(Node node, Canvas canvas, Paint paint, int line, String text) {
		paint.getTextBounds(text, 0, text.length(), stringBounds);
		paint.getFontMetrics(fontMetrics);
		float lineY = line * (fontMetrics.descent - fontMetrics.ascent + fontMetrics.leading) - fontMetrics.ascent;
		float totalHeight = fontMetrics.descent + fontMetrics.leading - 2 * fontMetrics.ascent;
		int paintedX = node.x;
		int width = node.width;
		if (width>1) {
			paintedX++;
			width--;
		}
		int paintedY = node.y;
		int height = node.height;
		if (height>1) {
			paintedY++;
			height--;
		}
		if (((width - 2) < (stringBounds.width())) || ((height - 2) < totalHeight)) {
			float minRatio = Math.min((width - 2.0f) / stringBounds.width(), (height - 2.0f) / totalHeight);
			paint.setTextSize(paint.getTextSize()*minRatio);
			paint.getFontMetrics(fontMetrics);
			lineY = line * (fontMetrics.descent - fontMetrics.ascent + fontMetrics.leading) - fontMetrics.ascent;
			totalHeight = fontMetrics.descent + fontMetrics.leading - 2 * fontMetrics.ascent;
			
		}
		canvas.save();
		canvas.clipRect(paintedX, paintedY, paintedX + width, paintedY + height);
		canvas.drawText(text, paintedX + 1, paintedY + lineY + 1, paint);    
		canvas.restore();
		return 1;
	}
}
